% coelho FEV 2003
\documentclass[language=brazil,structure=flat]{cweb}

%% Escrevendo em portugus:
\usepackage[brazil]{babel}
\usepackage[T1]{fontenc}
%----------------------------

\setlength{\topmargin}{-0.5in}
\setlength{\textheight}{9in}
\setlength{\oddsidemargin}{-.125in}
\setlength{\evensidemargin}{-.125in}
\setlength{\textwidth}{6.75in}

% Macros do LaTeX
\newcommand{\cweb}{\texttt{CWEB}}
\newcommand{\sgb}{\texttt{SGB}}
\newcommand{\defi}[1]{\textit{#1}}
\newcommand{\floor}[1]{\lfloor #1 \rfloor} 
\newcommand{\ceil}[1]{\lceil #1 \rceil}
\newcommand{\seq}[1]{\langle #1 \rangle}
\newcommand{\opcao}[1]{$\langle$#1$\rangle$}
\newcommand{\caminho}[1]{$\langle$#1$\rangle$}
\newcommand{\conjunto}[1]{\{\,#1\,\}}
\newcommand{\conj}[1]{\{#1\}}
\newcommand{\NP}{\textup{NP}}
\newcommand{\PP}{\textup{P}}
 
% \pch  problema do circuito hamiltoniano 
\newcommand{\pch}{\textsc{PCH}}

\newcommand{\gbb}{\textsc{gb\_basic}}
\newcommand{\gbr}{\textsc{gb\_rand}}
\newcommand{\gbs}{\textsc{gb\_save}}

\begin{document}

@i gb_types.w 

\title{{\small Exemplo de uso do Stanford GraphBase com \texttt{CWEB}-\LaTeX}\\
Grafos Hamiltonianos}
\author{coelho}

\maketitle

\tableofcontents

@*1 Introduo. Esta  uma demonstrao do \sgb\ com \cweb-\LaTeX.
Antes de descrever o problema e o programa para resolv-lo  conveniente 
introduzir alguma terminologia.

%%%%% Passeio
Um \defi{passeio} num grafo |g|  qualquer seqncia da forma
\begin{eqnarray}
\label{passeio}
\seq{v_{0}, a_{1}, v_{1}, \ldots, a_{k}, v_{k}}
\end{eqnarray}
onde $v_{0}, \ldots, v_{k}$ so vrtices, $a_{1}, \ldots, a_{k}$ 
so arcos e, para cada $i$, $a_{i}$  o arco $v_{i-1}v_{i}$. 
%%%%% incio e fim de passeios
O vrtice $v_{0}$  o \defi{incio} do passeio e o $v_{k}$  seu
\defi{trmino}.
%%%%% caminho 

%%%% caminho e circuito hamiltoniano
Um \defi{caminho}  um passeio
sem vrtices repetidos. J um \defi{circuito}  um caminho onde o incio
e trmino coincidem. 
Um caminho ou circuito que contm todos os vrtices do grafo  dito
\defi{hamiltoniano}.  
O \defi{problema do circuito hamiltoniano} (\textit{Hamiltonian circuit
  problem}), denotado pela sigla \pch\  o seguinte:

\begin{quote}
\textbf{Problema} \pch|(g)|: %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Dado um grafo |g|, encontrar um circuito hamiltoniano.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\end{quote}
 
Um grafo que possui um circuito hamiltoniano  dito \defi{hamiltoniano}.
O \pch\  um problema computacionalmente difcil.  Tecnicamente falando
o \pch\  \NP-difcil. Isto significa que todo algoritmo conhecido s
consegue resolver o problema para grafos `grandes' aps consumir uma
quantidade de tempo indecente.  \texttt{MAC0338} Anlise de Algoritmos 
uma disciplina que esclarece o significado desse termo de uma maneira
mais precisa.




@*1 O que o programa faz.  Este programa recebe um grafo, atravs de um
arquivo no formato do \sgb\ ou de uma opo para ger-lo, e decide se o grafo
dado  ou no hamiltomiano. 
 
Se o grafo  hamiltoniano o programa devolve, para
a sada padro, uma seqncia dos vrtices do grafo que formam um circuito
hamiltoniano. J, se o grafo no  hamiltoniano, o programa responde apenas
que lamenta \ldots Seria bacana se tivssemos um programa que 
devolvesse algum objeto que fosse uma prova cabal de que o grafo no 
hamiltoniano. Mais ainda, seria timo que o programa pudesse verificar
rapidamente essas provas ou certificados. Uma caracterstica dos problemas
\NP-difceis  que estes admitem um certificado `sucinto' (ou seja, verificvel
rapidamente) para a resposta \textsc{sim} (no caso, o grafo  hamiltoniano), mas no se
conhece um certificado sucinto para a respotas \textsc{no} (no caso, o grafo no 
hamiltoniano). 


@*1 Como usar o programa.
Para usar o programa o
usurio deve, na linha de comando, escrever
\begin{quote}
\texttt{meu\_prompt> hamiltoniano \opcao{opes}}
\end{quote}
onde \texttt{\opcao{opes}} consiste de zero ou mais das seguintes
especificaes em qualquer ordem
\begin{description}
\item[\texttt{-r}]  Usa a subrotina |random_graph| do mdulo
  \gbr\ do \sgb\ para gerar um grafo pseudo-aleatrio com |n|
  vrtice |m| arcos ou arestas, usando um gerador de nmeros
  pseudo-aleatrios que depende de um semente (veja opo \texttt{-s}). 

\item[\texttt{-c}] Usa a subrotina |circuit| do mdulo \gbb\ do
  \sgb\ para gerar um circuito simtrico com |n| vrtices. Todo
  circuito  hamiltoniano.

\item[\texttt{-e}] Usa a subrotina |empty| do mdulo \gbb\ do
  \sgb\ para gerar um grafo com |n| vrtices e zero arcos.
  O grafo sem arcos no  hamiltoniano.

\item[\texttt{-p}] Usa a subrotina |petersen| do mdulo \gbb\ do
  \sgb\ para gerar o grafo de Petersen. O grafo de Petersen no 
    hamiltoniano.

\item[\texttt{-t}] Usa a subrotina |transitive| do mdulo \gbb\ do
  \sgb\ para gerar um torneio com |n| vrtices. Um torneio  um grafo
  que tem um arco de |u| a |v| se |u < v|. Torneios no so hamiltonianos.

\item[\texttt{-k}] Usa a subrotina |complete| do mdulo \gbb\ do
  \sgb\ para gerar um grafo simtrico completo com |n| vrtices. Num
  grafo completo simtrico temos uma arco de |u| a |v| para cada para
  |u| e |v| de vrtices distintos. Todo grafo completo simtrico 
  hamiltoniano.

\item[\texttt{-g\opcao{nome}}] Nenhum grafo ser gerado. O grafo sobre
  o qual o programa trabalhar ser lido do arquivo  \opcao{nome}.

\item[\texttt{-s\opcao{nmero}}] Usa o nmero |s| como semente do
  gerardor de nmeros pseudo-aleatrios, onde |s|  o \opcao{nmero} dado.

\item[\texttt{-n\opcao{nmero}}] O grafo gerado ter |n| vrtices,
  onde |n|  o \opcao{nmero} dado.  

\item[\texttt{-m\opcao{nmero}}] O grafo gerado ter |m| arcos ou
  arestas, onde |m|  o \opcao{nmero} dado.

\item[\texttt{-h\opcao{nmero}}] Usa a heurstica do vrtice de grau de
  sada mnimo.

\item[\texttt{-o\opcao{nome}}] O grafo de trabalho  salvo no arquivo 
\opcao{nome}.
\end{description}
 evidente que s faz sentido usar apenas uma entre as opes \texttt{-r,
  -c, -e, -p, -t, -k, -g}. Tambm no faz sentido usar a opo
  \texttt{-g} junto com as opes \texttt{-n, -s} e \texttt{-m}. S faz
  sentido usar a opo \texttt{-m} junto com \texttt{-r}.
  
Como de hbito, caso o usurio escreva apenas 
\begin{center}
\texttt{meu\_prompt> hamiltoniano} 
\end{center}
o programa deve descrever o seu funcionamento.

@<Funo |mostre_uso|@>=
void 
mostre_uso()
{
 fprintf(stdout,@/
     "Uso: %s [-r|-c|-e|-p|-t|-k|-g<nome>] [-h] [-s<num>] [-n<num>] "
     "[-m<num>] [-o<nome>]\n"@/
     "     -r  grafo pseudo-aleatorio com n vertices e m arestas\n"@/
     "     -c  circuito com n vertices\n"@/
     "     -e  grafo com n vertices e zero arcos\n"@/
     "     -p  grafo de Petersen\n"@/
     "     -t  torneio com n vertices\n"@/
     "     -k  grafo simetrico completo com n vertices\n"@/
     "     -g<nome> trabalha com o grafo do arquivo <nome>\n"@/
     "     -s<num>  para o gerar numeros pseudo-aleatorios\n"@/
     "     -n<num> numero de vertices\n"@/
     "     -m<num> numero de arcos/arestas\n"@/
     "     -o<nome> grafo de trabalho sera salvo no arquivo <nome>\n"@/
                           , nome_programa);
 exit(0);
}

@*1 Estratgia de soluo. 
Para resolver certos problemas  necessrio enumerar todas as
configuraes (ou objetos) de um certo tipo. O nmero de configuraes 
tipicamente muito grande, e portanto os algoritmos enumerativos
usualmente consomem uma quantidade absurda de tempo. Este programa 
 uma implementao de uma algoritmo enumerativo do tipo
\textit{backtracking}. Este tipo de algoritmo tambm  conhecido como 
\defi{tipo fora-bruta}, pois examina sistematicamente todos os candidados a
soluo. Voc j estudou algoritmos enumerativos em 
\texttt{MAC0122}. Veja, por exemplo, o \defi{problema das $n$ rainhas}

\begin{center}
\texttt{http://www.ime.usp.br/\~{}coelho/mac122/conteudo/backtracking/nrainhas.c} 
\end{center}

\noindent
e do \defi{problema do escalonamento} 

\begin{center}
\texttt{http://www.ime.usp.br/\~{}coelho/mac122/eps/ep4/}~.
\end{center}

Algoritmos enumerativos so muitas vezes escritos de maneira
recursiva. Este programa, s por nos diverso, apresenta uma implementao
iterativa.

 

@*1 Estrutura geral do programa.
Abaixo encontra-se o `layout' do programa, como  visto pelo 
compilador \texttt{C}.


@p
#include <stdio.h>
#include <string.h>
#include "gb_graph.h" /* estruturas de dados padro do \sgb\ */
#include "gb_save.h"  /* rotinas para ler e salvar um grafo */ 
#include "gb_basic.h" /* rotinas para gerar alguns grafos bsicos */
#include "gb_rand.h"  /* produz um grafo pseudo-aleatrio */
@h@#
@<Variveis globais@>@;
@<Declarao de funes@>@;
int main (argc, argv)   
  int argc; /* o nmero de argumentos na linha de comando */
  char *argv[]; /* o vetor de strings contendo os argumentos */
{
   @<Variveis locais do |main|@>; 
   @<Examine as opes na linha de comando@>;
   @<Crie o grafo |g| dado@>;
   @<Verifique se |g|  hamiltoniano@>;
   if (salva_arquivo) @<Salve o grafo |g| no arquivo |salva_arquivo|@>;
   gb_recycle(g); /* devolve a memria usada por |g| */
   fclose(fpani);
   return 0; /* termino normal */
}




@  Gosto que |nome_programa| e algumas outras variveis sejam globais. 
Alm disso, se o programa encontra algum problema, ento um cdigo de erro 
devolvido atravs de |exit|.
 
@d nome_arco a.S /* o |nome| de um arco  armazenado no utility field |a| */ 

@<Variveis globais@>=
typedef enum {FALSE,TRUE} Boolean;
char *nome_programa = NULL; /* nome do programa */
enum { /* cdigo de erros */@/
  opo_invalida = -1,@/
  salva_grafo_erro = -2,@/
  putz_programa_errado = -3@/
};
 
@ A seguir est a lista de funes que fazem parte do programa. Hummm, neste
programa s temos uma funo, excluindo-se o |main|.
@<Declarao de funes@>=
@<Funo |mostre_uso|@>@;


@*1 Tratamento das opes.
Entre as vriaveis local do |main| teremos o grafo |g| sobre o qual
trabalharemos. Outras variveis so relativas a opes do programa. 
Estas vriveis so inicializadas com seu valores default. Assim
\begin{center}
\texttt{meu\_prompt> hamiltoniano -n30} 
\end{center}
tem o mesmo efeito que
\begin{center}
\texttt{meu\_prompt> hamiltoniano -r -n30 -m100} 
\end{center}
 
   
@<Variveis locais do |main|@>=
unsigned long n = 20;  /* nmero de vrtices */
unsigned long m = 100;  /* nmero de arcos    */

long directed = 1;     /* o grafo gerado no precisa ser simtrico */
long seed = 0;         /* semente para o gerador pseudo-aleatrio */

Boolean aleatrio  = FALSE;  /* gerar um grafo pseudo-aleatrio */
Boolean circuito   = FALSE; /* gerar um circuito  */
Boolean vazio      = FALSE; /* gerar um grafo sem arcos */
Boolean petersen   = FALSE; /* gerar o grafo de Petersen */
Boolean torneio    = FALSE; /* gerar um torneio */
Boolean completo   = FALSE; /* gerar um grafo completo */

Boolean heurstica = FALSE;  /* heurstica do grau de sada mnimo */
Boolean escolheu   = FALSE; /* indica se foi escolhido como criar o grafo */

char *nome_arquivo  = NULL;  /* nome do arquivo do grafo de trabalho */
char *salva_arquivo = NULL; /* para salvar o grafo de trabalho */
 
FILE *fpani;  /* arquivo com instrues de animao */
FILE *fpgraf; /* arquivo com o grafo no formato do Nelson */

@ As opes podem ser fornecidas em qualquer ordem. Se o grafo  dado
por mais de uma maneira (opes \texttt{-r, -c, -e, -p, -t, -k} ou
\texttt{-g}), o programa pra com uma mensagem de erro. A varivel
|escolheu|  usada para detectar este tipo de inconsistncia. Est  a
nica inconsistncia verificada na linha de comando. Assim, por exemplo, 
o programa no reclama se for chamado com as opes \texttt{-c -n10 -m1000};
  note que todo circuito com 10 vrtices tem 20 arcos. 


@<Examine as opes ...@>=
nome_programa = argv[0]; 
if (argc == 1) mostre_uso();
while (--argc) {
  if      (strcmp(argv[argc],"-r")==0 && !escolheu) escolheu = aleatrio  = TRUE;
  else if (strcmp(argv[argc],"-c")==0 && !escolheu) escolheu = circuito   = TRUE;
  else if (strcmp(argv[argc],"-e")==0 && !escolheu) escolheu = vazio      = TRUE;
  else if (strcmp(argv[argc],"-p")==0 && !escolheu) escolheu = petersen   = TRUE;
  else if (strcmp(argv[argc],"-t")==0 && !escolheu) escolheu = torneio    = TRUE;
  else if (strcmp(argv[argc],"-k")==0 && !escolheu) escolheu = completo   = TRUE;
  else if (strncmp(argv[argc],"-g",2)==0 && !escolheu) { escolheu = TRUE; 
                                       nome_arquivo=argv[argc]+2;}
  else if (strcmp(argv[argc],"-h")==0) heurstica = TRUE;
  else if (sscanf(argv[argc],"-n%lu",&n)==1)    ;
  else if (sscanf(argv[argc],"-m%lu",&m)==1)    ;
  else if (sscanf(argv[argc],"-s%ld",&seed)==1) ;
  else if (strncmp(argv[argc],"-o",2)==0) salva_arquivo=argv[argc]+2;
  else {
    fprintf(stderr,
       "Uso: %s [-r|-c|-e|-p|-t|-k|-g<nome>]] [-h] [-s<semente>] [-n<num>] [-m<num>] [-o<nome>]\n",
        nome_programa);
    exit(opo_invalida);
  }

}
if (!escolheu) aleatrio = TRUE; /* opo default */
 


@*1 Criao do grafo. O trabalho desta parte do programa  construir ou
reconstruir o grafo |g| sobre o qual trabalharemos. O grafo ser
construdo se algumas das opes de gerao automtica do grafo foi
usada. Caso a opo \texttt{-g\opcao{nome}} tenha sido usada o grafo
ser recontrudo a partir do contedo de |nome_arquivo|. A criao
do grafo depende da

@<Variveis locais do |main|@>=
Graph *g = NULL; /* grafo de trabalho */
long v_num = 0;
long a_num = 0;

@ Se uma rotina de criao do grafo encontrar um problema, ela devolve
|NULL| (ou seja, \texttt{NULL}), depois de colocar um nmero de cdigo que
identifica o erro na varivel externa |panic_code|.  Se no ocorrer
problemas a rotina devolve o endereo do grafo criado (|Graph *|).


@<Crie o grafo |g| dado@>=
if (aleatrio)     g = random_graph(n, m, 0, 0, directed, NULL, NULL, 1, 1, seed);
else if (circuito) g = circuit(n);
else if (vazio)    g = empty(n);
else if (petersen) g = petersen();
else if (torneio)  g = transitive(n);
else if (completo) g = complete(n);
else if (nome_arquivo) g = restore_graph(nome_arquivo);
@/
if (g == NULL)@/
{@/
fprintf(stderr, "%s: grafo nao pode ser criado (codigo do erro: %ld)\n",
            nome_programa, panic_code);
return -1; /* Opsss ... */
}

fprintf(stdout,"grafo: %s.\nnumero de vertices: %ld.\nnumero de arcos: %ld.\n", 
                g->id, g->n, g->m);
 
fpani = fopen("nelson3.gaf","w"); 
fprintf(fpani,"LoadGraph(\"nelson3.grb\");\n"); 
 
fpgraf = fopen("nelson3.grb","w");
fprintf(fpgraf,"GRABBIT-Graph(%ld,%ld,1)\n", g->m, g->n);
for (u=g->vertices; u < g->vertices+g->n; u++, v_num++) 
  { 
    char nome[40];

    sprintf(nome,"V%ld",v_num);
    u->name=gb_save_string(nome); /* essa funo aloca memria para o
                                     nome do vrtice */
    fprintf(fpgraf,"%s,%ld,30,\"%2ld\",(255,0,0)\n",nome,30*v_num,v_num);
  }
 
for (u=g->vertices; u < g->vertices+g->n; u++) 
  {
     Arc *a;
 
     for (a = u->arcs; a != NULL; a=a->next, a_num++)
       {
         char nome[40];
  
         sprintf(nome,"A%ld",a_num);
         fprintf(fpgraf,"%s,\"%2ld\",(0,0,0),%s,%s\n",
                                 nome,a_num,u->name, a->tip->name);
         a->nome_arco=gb_save_string(nome);         
       }
  }
 
fclose(fpgraf);

@*1 Algoritmo enumerativo. Finalmente chegamos no corao do programa.
Durante a execuo do programa cada vrtice do grafo est no estado `ativo' ou
`inativo'. Os vrtices ativos fazem parte de um caminho na iterao corrente
de um vrtice |s| a um vrtice |u| e
os vrtices inativos so os que no esto no caminho. Grosseiramente falando,
a misso da iterao  estender esse caminho corrente a um caminho
hamiltoniano de |s| a um certo vrtice |t| ou se arrepender-se
(\textit{backtrack}) e voltar nas suas `prprias pegadas'.
 
A estruturas de dados que representa o estado corrente do algoritmo 
implementada usando trs `utility fields' em cada vrtice: |estado|, |pred|
(predecessor) e |current_arc|.  Consideraremos cada um deles separadamente.


@ O |estado| de um vrtice  um inteiro. Ele  zero (ou, |ATIVO|) quando o
vrtice est inativo. Assim que o vrtice passa a ser a ponta final |u| do
caminho corrente ele se torna ativo e o valor de seu |estado| passa a ser um
(ou |ATIVO|).  Ao final do programa, caso um caminho hamiltoniano de |s| a |t|
tenha sido encontrado, o programa certifica-se do seu achado e durante este
processo transforma o estado dos vrtices de |ATIVO| para |MARCADO|.
 
@d estado z.I /* o |estado| de um vrtice  armazenado no utility field |z| */
 
@<Variveis globais@>=
enum {INATIVO = 0, ATIVO = 1, MARCADO = 2};
 
@ Os vrtices ativos formam um caminho de um vrtice inicial |s| a um vrtice
|u|.  Todo vrtice ativo tem um predecessor no caminho, indicado pelo campo
|pred|. Este predecessor  outro vrtice
ativo; a nica excesso  a ponta inicial |s| do caminho que tem como 
predecessor o vrtice especial |t|, que no  ativo. Assim, 
um arco de |v| a |w| no caminho 
representado na estrutura por |w->pred == v|. O campo |pred| permite que o
programa faa backtracking no caminho. 
 
Assim, no fundo, o programa mantm uma representao `reversa' do caminho
corrente de |s| a |u|,  definida pelos campos |pred| de cada
vrtice ativo, ou seja, o caminho corrente 
\begin{center}
\caminho{|s=u->pred->|\ldots|->pred, |\ldots|, u->pred->pred->pred,
u->pred->pred, u->pred, u|}.
\end{center}
 
@d pred y.V /* o predecessor (|pred|) de um vrtice e armazenado no utility
field |y|*/
 
@<Variveis locais ...@>=
Vertex *s   ; /* vrtice que  ponta inicial do caminho corrente */
Vertex *u   ; /* vrtice que  ponta final do caminho corrente */
Vertex *t   ; /* vrtice destino do caminho corrente */
 
 
@ Todos os arcos do grafo so explorados sistematicamente, a partir do vrtice
origem |s|, como numa busca em profundidade. Suponha que no caminho corrente
existe um arco de |v| a |w|. Logo, |w->pred = v|. O arco de |v| a |w| tambm 
representado diretamente na estrutura pelo campo |current_arc| do
vrtice |v|. Desta forma,  tem-se que |v->current_arc->tip = w|. 
Este tipo de apontador  bem conhecido e til (veja
o livro \textit{Network Flows: Theory, Algorithms, and Applications},
R.K.~Ahuja, Th.L.~Magnanti, J.B.~Orlin, Prentice Hall, 1993). 
O caminho corrente  representado por
\begin{center}
\caminho{|s, s->current_arc->tip,s->current_arc->tip->current_arc->tip,
|\ldots|, s->current_arc->tip->|\ldots|->current_arc->tip, u|}.
\end{center}
 
@d current_arc x.A 
 

@ Busca em profundidade explora um grafo sistematicamente, visitando
todos os vrtices e verificando para onde estes vrtices podem nos conduzir.   
O vrtice ativo que  ponta final do caminho corrente  |u|.
 
H duas possibilidade, o arco |u->current_arc| tenha ponta final em um
vrtice |v| que   ativo ou inativo.  
Se o vrtice |v|  ativo, ento |v| j faz parte do caminho corrente e
tentaremos estender o caminho atravs do arco sucessor de
|u->current_arc| na lista de adjacncia de |u|. Caso |v| seja inativo, 
marcamos |v| passa a ser a nova ponta final e do caminho corrente e passamos 
a examinar
os arcos com ponta inicial nele. Se chegarmos a um ponto em que todos os
arcos com ponta inicial em |u| nos levam a vrtices ativos, chegou a
hora de dar um passo atrs e fazer com que |u->pred| volte a ser a ponta final
do caminho.
 
Para tentar encontrar um circuito hamiltoniano no grafo |g|
precederemos da seguinte maneira. O \defi{grau de sada} de um vrtice 
o nmero de arcos que tm ponta inicial nele. 
Inicialmente o programa escolhe um vrtice `destino' |t|;
se a opo \texttt{-h} foi usada escolheremos como |t| um vrtice de |g|
que tem grau de sada mnimo. Se
encontramos em |g| um caminho hamiltoniano de |s| a |t|, ento |g| 
hamiltoniano.  De fato, a concatenao do caminho ao arco de |t| a |s|  um
circuito hamiltoniano. 
 
A idia bsica do algoritmo enumerativo abaixo  a seguinte. Com |t|
fixado tente, para cada arco da forma |ts|, encontar um caminho
hamiltoniano de |s| a |t|.  evidente que se todas as buscas por um tal
caminho falharem, ento o grafo no  hamiltoniano.

@ O campo |util_types| de um grafo contm um string de comprimento 14,
seguido do caractere nulo `|\0|'. Os primeiros 6 caracteres do
|util_type| indicam o uso dos campos utilitrios |u, v, w, x, y| e |z|
dos registros |Vertex|. Os prximos dois caracteres do o formato dos
campos utilitrios |a| e |b| em |Arc|. Finalmente, os seis ltimo caracteres
mostram o formato dos campos utilitrios |uu, vv, ww, xx, yy| e |zz| em 
|Graph|.
Cada caractere deve ser um \texttt{I} (denotando um inteiro longo),
\texttt{S}  (representando o endereo de um string), \texttt{V}
(indicando o endereo de um |Vertex|), \texttt{A} (denotando o endereo
de um |Arc|), \texttt{G} (endereo de um |Graph|), ou \texttt{Z}
(significando que o campo no est em uso).

@<Crie ...@>=
  strcpy(g->util_types,"ZZZAVIZZZZZZZZ"); 
    
@ Estrutura geral do algoritmo enumerativo.
Seja |t| um vrtice de |g|.  evidente que se |g|  hamiltoniano ento existe
um arco, digamos de |t| a |s|, que faz parte desse circuito. Logo, existe um
caminho hamiltoniano de |s| a |t| em |g|. Este fato determina o comportamento
do trecho de cdigo a seguir: um vrtice |t| e determinado e para cada arco da
forma |ts| o programa procura um caminho hamiltoniano de |s| a |t|.
Se para algum arco |ts| um tal caminho  encontrado enta o grafo 
hamiltoniano, caso contrrio o programa declara que o grafo dado no 
hamiltoniano e pra.
    
@<Verifique se |g|  hamiltoniano@>= 
{@/
  Boolean hamiltoniano = FALSE;@#
 
  if (heurstica)@/ 
     @<Encontre em |g| um vrtice |t| de grau de sada mnimo@>@;
  else  t = g->vertices; 
 

  @<Marque todos vrtices como inativos@>;
 
  fprintf(fpani,"PaintVertex(%s,(0,0,255));\n", t->name);
 
  t->current_arc = t->arcs;  
  while (t->current_arc && !hamiltoniano)@/ 
  {@/  
    fprintf(fpani,"PaintArc(%s,(0,0,255));\n", t->current_arc->nome_arco);
    s = t->current_arc->tip;
    s->pred = t;
    @<Se existe um caminho hamiltoniano de |s| a |t|, |hamiltoniano = TRUE|@>;
    if (!hamiltoniano) 
      {@/
          fprintf(fpani,"PaintArc(%s,(0,0,0));\n", 
                                        t->current_arc->nome_arco);  
          t->current_arc = t->current_arc->next;
      }
  } 
  if (hamiltoniano)@/
    {@/
    @<Verifique se os |current_arc|'s formam um circuito hamiltoniano@>;
    fprintf(stdout, "O grafo e hamiltoniano.\n");
    }
  else 
    fprintf(stdout, "O grafo nao e hamiltoniano.\n");
 
  fprintf(stdout,"no backtrackings: %ld.\nno nos gerados: %ld.\n",
                    no_backtrackings, no_nos);
}

@ 
@<Marque todos ...@>=
{ 
 register Vertex *v;
 
 for (v=g->vertices; v < g->vertices+g->n; v++)
   {@/    
    v->estado = INATIVO;
    fprintf(fpani,"PaintVertex(%s,(255,255,255));\n", v->name);
   }
}
 
@ Ao final deste bloco a varivel |t| contm um vrtice
em |g| de grau de sada mximo.   
 
@<Encontre em |g| um vrtice ...@>=
{
  register Vertex  *v;
  register Arc *a; 
  unsigned long grau_min = 0;

  t = g->vertices;
  for (a=t->arcs; a; a=a->next) grau_min++;
 
  for (v=g->vertices+1; v < g->vertices+g->n; v++)@/ 
   {@/  
     unsigned long grau = 0;
     for (a=v->arcs; a; a=a->next) grau++;
     if (grau < grau_min)@/ 
       {@/
          t = v;
          grau_min = grau;
       }        
   }        

  fprintf(stdout, "Vertice origem: %s.\nGrau de saida: %lu.\n", 
           t->name, grau_min);
}
 

@ Cada iterao do trecho de cdigo deste bloco comea com um caminho de |s| a
|u|. Cada iterao tenta estender este caminho a um caminho hamiltoniano que
v de |s| a |t|. Assim, por exemplo, se no incio de uma iterao tem-se que 
|u=t| e o comprimento do caminho  |g->n-1|, tem-se que um caminho
hamiltoniado de |s| a |t| acaba de ser encontrado. 
 
O nmero de iteraes do lao |while (!tentou_tudo && !hamiltoniano)|  
essencialmente dado pela soma de |no_nos| e |no_backtrackings|. A vrivel
|no_nos| indica o nmero de ns gerados na rvore de deciso e |no_backtrackings|
indica o nmero de arrependimentos feitos pelo programa.

Durante a execuo o nmero de ns gerados da rvore de deciso 
contado. A cada 10 Mega ns um ponto (`.')  impresso para indicar que o
programa ainda est trabalhando.
 

@<Se existe ...@>= 
{
  unsigned long comprimento; /* nmero de arcos no caminho corrente */  
  Boolean tentou_tudo; /* indica se todos os candidatos a soluo foram
                            testados */ @#
  u = s;
  fprintf(fpani,"PaintVertex(%s,(255,0,0));\n", u->name);
  u->current_arc = u->arcs;
 
  if (u->current_arc != NULL)
     fprintf(fpani,"PaintArc(%s,(0,0,255));\n", u->current_arc->nome_arco);
  u->estado = ATIVO;
  comprimento = 0;
  tentou_tudo = FALSE;
  while (!tentou_tudo && !hamiltoniano) @/
    {@/
       if (u == t && comprimento == g->n - 1)
         hamiltoniano = TRUE;
       else if (u == s && u->current_arc == NULL)
         {
            tentou_tudo = TRUE;
            u->estado = INATIVO;
            fprintf(fpani,"PaintVertex(%s,(255,255,255));\n", u->name);
         }
       else if (u == t || u->current_arc == NULL) @\ 
          { /* backtracking */ @\
              no_backtrackings++; 
              u->estado = INATIVO;
              fprintf(fpani,"PaintVertex(%s,(255,255,255));\n", u->name);
              comprimento--;
              fprintf(fpani,"PaintArc(%s,(0,0,0));\n", 
                                    u->pred->current_arc->nome_arco);
              u = u->pred;
              fprintf(fpani,"PaintVertex(%s,(255,0,0));\n", u->name);
              u->current_arc = u->current_arc->next;
              if (u->current_arc != NULL)
              fprintf(fpani,"PaintArc(%s,(0,0,255));\n", 
                                 u->current_arc->nome_arco);
          }
       else if (u->current_arc->tip->estado == ATIVO)@/
         { /* vamos tentar o prximo arco */ @/ 
           no_nos++;
           fprintf(fpani,"PaintArc(%s,(0,0,0));\n", 
                                 u->current_arc->nome_arco);
           u->current_arc = u->current_arc->next;
           if (u->current_arc != NULL)
             fprintf(fpani,"PaintArc(%s,(0,0,255));\n", 
                                 u->current_arc->nome_arco);
         }
       else @/
         {/* cresce o caminho corrente */ @/ 
           Vertex *v;@#

           no_nos++; @/
           fprintf(fpani,"PaintVertex(%s,(0,0,255));\n", u->name);
           v = u->current_arc->tip;
           v->estado = ATIVO;
           
           comprimento++;
           v->pred = u;
           if (v != t)
             { 
                fprintf(fpani,"PaintVertex(%s,(255,0,0));\n", v->name); 
                v->current_arc = v->arcs;
                if (v->current_arc != NULL) 
                fprintf(fpani,"PaintArc(%s,(0,0,255));\n", 
                                 v->current_arc->nome_arco);
             }
           u = v;
           }
       if (no_nos > 0 && no_nos % 10000000 == 0) /* s para ver se o prog. ainda est vivo */@/ 
         { 
            fprintf(stdout,"."); 
            fflush(stdout);
         }

     }@/

   /* fprintf(stdout,"\n"); */

}


@ 

@<Variveis locais ...@>=
unsigned long no_nos = 0;
unsigned long no_backtrackings = 0;


@ Se o programa encontra um circuito hamiltoniano, ento os
|current_arc|'s formam este circuito. Neste bloco so contados
e marcados todos os vrtices que fazem parte do suposto circuito
hamiltoniano. Caso os |current_arc|'s no formem um circuito
hamiltoniano,  o programa para aps escrever uma mensagem de erro.
 

@<Verifique se os ...@>=
{@/
  register Arc *a;
  long comprimento; /* nmero de arcos percorridos */@#
  
  fprintf(stdout,"Circuito hamiltoniano: "); 
  fprintf(stdout,"%s", t->name);
  for (a = t->current_arc, comprimento = 1; a->tip != t; a = a->tip->current_arc)
    {@/
       fprintf(stdout,", %s", a->tip->name);     
       if (a->tip->estado == ATIVO) 
         {@/
           comprimento++;
           a->tip->estado = MARCADO;
         }
     }@#
 
   fprintf(stdout,"\n");
   if (comprimento != g->n)
     {
       fprintf(stderr,"%s: os current_arc's nao formam um circuito hamiltoniano!\n"
                      "comprimento: %ld.\n",  
       nome_programa, comprimento);
       return (putz_programa_errado); 
     }
}

@*1 Salva grafo. Se o usurio utilizou a opo 
\texttt{-o}\opcao{nome} o grafo |g|  salvo no arquivo 
\opcao{nome}.  usual que o arquivo que contm um grafo no formato do
\sgb\ tenha a extenso \texttt{.gb}. Para salvar um grafo usamos a
subrotina |gb_save| do mdulo \gbs\ do \sgb.



@<Salve ...@>=
{@/
long anomalia = save_graph(g, salva_arquivo);
if (anomalia)@/ 
  {@/
    fprintf(stderr, "%s: grafo nao pode ser salvo no arquivo"
                    " %s (codigo do erro: %ld)\n",
             nome_programa, salva_arquivo, anomalia);
    gb_recycle(g);
    exit(salva_grafo_erro); /* Opsss ... */
  }
}
 
@

@*1 Exerccios. 
\begin{enumerate}
\item[1.]  Qual o consumo de tempo do programa quando |g|  um
 circuito com |n| vrtices 
 (opo \texttt{-c})? A heurstica (opo \texttt{-h}) reduz o consumo
 de tempo neste caso?

\item[2.]  Qual o consumo de tempo do programa quando |g|  um
 grafo com |n| vrtices e sem arcos
 (opo \texttt{-e})? A heurstica (opo \texttt{-h}) reduz o consumo
 de tempo neste?


\item[3.] Qual o consumo de tempo do programa quando |g|  um
  torneio com |n| vrtices 
 (opo \texttt{-t})? A heurstica (opo \texttt{-h}) reduz o consumo
 de tempo neste caso?

\item[4.] Qual o consumo de tempo do programa quando |g|  um
 grafo completo com |n| vrtices 
 (opo \texttt{-c})? A heurstica (opo \texttt{-h}) reduz o consumo
 de tempo neste caso?
 
\item[5.] Qual o consumo de tempo assinttico no pior caso deste
  programa? Apresente um tipo de grafo no qual este consumo de tempo 
  alcanado. 

 
\item[6.] Descreva alguns tipos de grafos para os quais a heurtica do 
vrtice de menor de grau de sada (opo \texttt{-h}) faz com que o
programa/algoritmo gaste uma quantidade de tempo significativamente menor.

\item[7.] Descreva um exemplo em que o tempo gasto pelo
  programa/algoritmo utilizando a heurstica  maior que se no ela no
  fosse usada.

\item[8.] `Envenene' esta implementao atravs de mais algumas
  heursticas.

\item[9.] O \sgb\ possui mais geradores de grafos. Acrescente esses
  geradores a este programa e considere o consumo de tempo do programa
  tende esses grafos como entrada.

\item[10.] Algoritmos de enumerao  so muitas vezes escritos de
  manera recursiva. Quais estruturas de dados que na presente
  implementao fazem o papel da tal pilha de recurso?
 
\item[11.] Escreva uma funo recursiva 
\begin{center}
|Boolean (Graph *g){...}|
\end{center}
que devolve |TRUE| se |g|  hamiltoniano e |FALSE| caso contrrio.

\end{enumerate}

\cwebIndexIntro{%
    Aqui vai uma lista de identificadores usados e onde eles aparecem.
Entradas grifadas indicam os lugares de definies. Mensagens de erros
    tambm so mostradas.
}

\end{document}


